﻿using System;
using Amqp;
using Amqp.Framing;

namespace AmqpClient
{
    /// <summary>
    /// Base abstract class for AMQP client (sender and receiver)
    /// </summary>
    public abstract class AmqpClientEntity : IDisposable
    {
        protected bool disposed = false;

        // broker address
        private Address address;
        // connection with the broker
        protected Connection connection;
        // session of current connection
        protected Session session;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="host">Broker host URI</param>
        /// <param name="port">Broker port</param>
        /// <param name="user">Username</param>
        /// <param name="password">Password</param>
        public AmqpClientEntity(string host, int port, string user, string password) 
            : this(new Address(host, port, user, password))
        {

        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="address">Address instance</param>
        public AmqpClientEntity(Address address)
        {
            this.address = address;

            // create and establish connection and session
            this.Init();
        }

        /// <summary>
        /// Initialize connection and session
        /// </summary>
        private void Init()
        {
            // create and establish connection
            this.connection = new Connection(this.address);
            this.connection.OnClosed += this.ConnectionClosed;

            // create a session
            this.session = new Session(this.connection);
            this.session.OnClosed += this.SessionClosed;
        }

        protected abstract void ConnectionClosed(AmqpObject sender, Error error);
        
        protected abstract void SessionClosed(AmqpObject sender, Error error);
        
        public virtual void Close()
        {
            this.Dispose(true);
        }

        #region IDisposable ...

        /// <summary>
        /// Destructor
        /// </summary>
        ~AmqpClientEntity()
        {
            // the object went out of scope and finalized is called
            // call dispose ("false" param) to release unmanaged resources 
            // NOTE : the managed resources will anyways be released when GC 
            //        runs the next time.
            Dispose(false);
        }


        public void Dispose()
        {
            // method called from user that wants to release all resources (managed and not)
            Dispose(true);

            // we have already cleaned up so the finalizer is useless
            // (tell the GC not to call it later)
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Dispose object resources
        /// </summary>
        /// <param name="disposing">Dispose managed resources or not</param>
        protected virtual void Dispose(bool disposing)
        {
            if (this.disposed)
                return;

            if (disposing)
            {
                // we want explicitily clean up resources (managed)
                this.session.Close();
                this.connection.Close();
            }
            else
            {
                // disposing "false" means that object went out of scope
                // finalizer is called and GC will release resources on next run
            }

            // release unmanaged resource because they will not be released by GC
            
            this.disposed = true;
        }

        #endregion
    }
}
