﻿using Amqp;
using Amqp.Framing;

namespace AmqpClient
{
    /// <summary>
    /// AMQP Receiver class
    /// </summary>
    public class AmqpReceiver : AmqpClientEntity
    {
        /// <summary>
        /// Delegate for message received callback (via message pump)
        /// </summary>
        /// <param name="receiver">AmqpReceiver instance</param>
        /// <param name="message">Message received</param>
        public delegate void AmqpMessageCallback(AmqpReceiver receiver, Message message);

        // default value for credit
        private const int DEFAULT_CREDIT = 5;

        // link with remote node (queue, topic, subscription)
        private ReceiverLink link;
        // entity node path
        private string entity;

        // credit for receiving
        private int credit;
        // callback for message pump
        private AmqpMessageCallback OnAmqpMessage;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="host">Broker host URI</param>
        /// <param name="port">Broker port</param>
        /// <param name="user">Username</param>
        /// <param name="password">Password</param>
        /// <param name="entity">Entity node path</param>
        public AmqpReceiver(string host, int port, string user, string password, string entity) 
            : this(new Address(host, port, user, password), entity)
        {

        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="address">Address instance</param>
        /// <param name="entity">Entity node path</param>
        public AmqpReceiver(Address address, string entity)
            : base(address)
        {
            this.entity = entity;

            this.Init();
        }

        /// <summary>
        /// Initialize link with entity node
        /// </summary>
        private void Init()
        {
            // create link with entity node
            this.link = new ReceiverLink(this.session, "receive-link " + this.entity, this.entity);
            this.link.OnClosed += this.LinkClosed;

            // set credit for receiving
            this.link.SetCredit(this.credit);
        }

        /// <summary>
        /// Set credit for receiving messages
        /// </summary>
        /// <param name="credit">Credit</param>
        public void SetCredit(int credit)
        {
            this.link.SetCredit(credit);
        }

        /// <summary>
        /// Receives a message
        /// </summary>
        /// <returns>Message received (null if no message available)</returns>
        public Message Receive()
        {
            return this.link.Receive();
        }

        /// <summary>
        /// Completes processing of a message
        /// </summary>
        /// <param name="message">Message to complete processing</param>
        public void Complete(Message message)
        {
            this.link.Accept(message);
        }

        /// <summary>
        /// Discards the message and relinquishes the message lock ownership
        /// </summary>
        /// <param name="message">Message to discard and unlock</param>
        /// <param name="reject">Reject or not the message</param>
        public void Abandon(Message message, bool reject)
        {
            if (reject)
                this.link.Reject(message);
            else
                this.link.Release(message);
        }

        /// <summary>
        /// Discards the message and relinquishes the message lock ownership
        /// </summary>
        /// <param name="message">Message to discard and unlock</param>
        public void Abandon(Message message)
        {
            this.Abandon(message, false);
        }

        /// <summary>
        /// Processes a message in an event-driven message pump
        /// </summary>
        /// <param name="OnAmqpMessage">The method to invoke when the operation is complete</param>
        /// <param name="credit">Credit for receiving messages</param>
        public void OnMessage(AmqpMessageCallback OnAmqpMessage, int credit)
        {
            this.OnAmqpMessage = OnAmqpMessage;
            this.credit = credit;

            this.link.Start(this.credit, this.OnMessageCallback);
        }

        /// <summary>
        /// Processes a message in an event-driven message pump
        /// </summary>
        /// <param name="OnAmqpMessage">The method to invoke when the operation is complete</param>
        public void OnMessage(AmqpMessageCallback OnAmqpMessage)
        {
            this.OnMessage(OnAmqpMessage, DEFAULT_CREDIT);
        }

        private void OnMessageCallback(ReceiverLink receiver, Message message)
        {
            AmqpMessageCallback callback = this.OnAmqpMessage;

            // message received from AMQP low level, raise event for AMQP client user
            if (callback != null)
                callback(this, message);
        }

        private void LinkClosed(AmqpObject sender, Error error)
        {
            this.link = null;
        }

        protected override void ConnectionClosed(AmqpObject sender, Error error)
        {
            this.connection = null;
        }

        protected override void SessionClosed(AmqpObject sender, Error error)
        {
            this.session = null;
        }

        #region IDisposable ...

        /// <summary>
        /// Destructor
        /// </summary>
        ~AmqpReceiver()
        {
            // the object went out of scope and finalized is called
            // call dispose ("false" param) to release unmanaged resources 
            // NOTE : the managed resources will anyways be released when GC 
            //        runs the next time.
            Dispose(false);
        }

        protected override void Dispose(bool disposing)
        {
            if (this.disposed)
                return;

            if (disposing)
            {
                // we want explicitily clean up resources (managed)
                this.link.Close();
            }
            else
            {
                // disposing "false" means that object went out of scope
                // finalizer is called and GC will release resources on next run
            }

            // release unmanaged resource because they will not be released by GC

            base.Dispose(disposing);
        }

        #endregion
    }
}
