﻿using Amqp;
using Amqp.Framing;

namespace AmqpClient
{
    /// <summary>
    /// AMQP Sender class
    /// </summary>
    public class AmqpSender : AmqpClientEntity
    {
        // link with remote entity node (queue, topic, subscription)
        private SenderLink link;
        // entity node path
        private string entity;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="host">Broker host URI</param>
        /// <param name="port">Broker port</param>
        /// <param name="user">Username</param>
        /// <param name="password">Password</param>
        /// <param name="entity">Entity node path</param>
        public AmqpSender(string host, int port, string user, string password, string entity) 
            : this(new Address(host, port, user, password), entity)
        {

        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="address">Address instance</param>
        /// <param name="entity">Entity node path</param>
        public AmqpSender(Address address, string entity) 
            : base(address)
        {
            this.entity = entity;

            this.Init();
        }

        /// <summary>
        /// Initialize link with entity node
        /// </summary>
        private void Init()
        {
            // create link with entity node
            this.link = new SenderLink(this.session, "send-link " + this.entity, this.entity);
            this.link.OnClosed += this.LinkClosed;
        }

        /// <summary>
        /// Send a message
        /// </summary>
        /// <param name="message">Message to send</param>
        public void Send(Message message)
        {
            this.link.Send(message, null, null);
        }

        public override void Close()
        {
            base.Close();

            this.Dispose(true);
        }

        private void LinkClosed(AmqpObject sender, Error error)
        {
            this.link = null;
        }

        protected override void ConnectionClosed(AmqpObject sender, Error error)
        {
            this.connection = null;
        }

        protected override void SessionClosed(AmqpObject sender, Error error)
        {
            this.session = null;
        }

        #region IDisposable ...

        /// <summary>
        /// Destructor
        /// </summary>
        ~AmqpSender()
        {
            // the object went out of scope and finalized is called
            // call dispose ("false" param) to release unmanaged resources 
            // NOTE : the managed resources will anyways be released when GC 
            //        runs the next time.
            Dispose(false);
        }

        protected override void Dispose(bool disposing)
        {
            if (this.disposed)
                return;

            if (disposing)
            {
                // we want explicitily clean up resources (managed)
                this.link.Close();
            }
            else
            {
                // disposing "false" means that object went out of scope
                // finalizer is called and GC will release resources on next run
            }

            // release unmanaged resource because they will not be released by GC

            base.Dispose(disposing);
        }

        #endregion
    }
}
