﻿using System;
using System.Collections;
using Amqp;
using Amqp.Framing;
using AmqpClient;

namespace IoT.Device.CommLayer
{
    /// <summary>
    /// AMQP communication layer
    /// </summary>
    public class AmqpCommLayer : ICommLayer
    {
        // AMQP sender and receiver instances
        private AmqpSender sender;
        private AmqpReceiver receiver;
        
        #region ICommLayer ...

        public void Start(string deviceId, string deviceKey, string hostUri, string egressUri, string ingressUri, string cmdIngressUri)
        {
            // create connection string and address to access Device Gateway via AMQP on Service Bus
            string connString = String.Format("amqps://{0}:{1}@{2}", deviceId, deviceKey, hostUri);
            Address address = new Address(connString);

            // create AMQP client sender and receiver
            this.sender = new AmqpSender(address, ingressUri);
            this.receiver = new AmqpReceiver(address, egressUri);

            // set callback for received messages
            this.receiver.OnMessage(this.OnMessageCallback);
        }

        public void Stop()
        {
            this.sender.Close();
            this.receiver.Close();
        }

        public void SendTelemetry(string deviceId, string subject, uint id, object value)
        {
            Message message = new Message();

            message.Properties = new Properties();
            message.Properties.Subject = subject;

            message.ApplicationProperties = new ApplicationProperties();
            message.ApplicationProperties["From"] = deviceId;
            message.ApplicationProperties["Time"] = DateTime.UtcNow;
            message.ApplicationProperties["dev"] = deviceId;
            message.ApplicationProperties["id"] = id;
            message.ApplicationProperties["val"] = value;

            sender.Send(message);
        }

        public void SendCommandReply(string deviceId, string gesture, string relatesTo, object result)
        {
            Message message = new Message();

            message.Properties = new Properties();
            message.Properties.Subject = gesture;

            message.ApplicationProperties = new ApplicationProperties();
            message.ApplicationProperties["From"] = deviceId;
            message.ApplicationProperties["Time"] = DateTime.UtcNow;
            message.ApplicationProperties["RelatesTo"] = relatesTo; // correlation with command
            message.ApplicationProperties["res"] = result;

            sender.Send(message);
        }

        public event CommandReceivedCallback OnCommandReceived;

        #endregion

        private void OnMessageCallback(AmqpReceiver receiver, Message message)
        {
            try
            {
                if (message != null)
                {
                    IDictionary properties = new Hashtable();

                    // add gesture
                    properties.Add("gst", message.Properties.Subject);
                    // add messageId
                    properties.Add("messageId", message.Properties.MessageId);
                    
                    // add all other properties
                    foreach (DictionaryEntry entry in message.ApplicationProperties)
                    {
                        properties.Add(entry.Key, entry.Value);
                    }

                    // raise command received event
                    CommandReceivedCallback callback = this.OnCommandReceived;

                    if (callback != null)
                    {
                        bool result = callback(this, properties);

                        if (result)
                            // message "accepted"
                            receiver.Complete(message);
                        else
                            // message "rejected" (invalid and/or unprocessable)
                            receiver.Abandon(message, true);
                    }
                }
            }
            catch
            { }
        }
    }
}
