﻿using System;
using System.Collections;
using System.Text;
using uPLibrary.Networking.M2Mqtt;
using uPLibrary.Networking.M2Mqtt.Messages;
using Cet.MicroJSON;

namespace IoT.Device.CommLayer
{
    /// <summary>
    /// MQTT communication layer
    /// </summary>
    public class MqttCommLayer : ICommLayer
    {
        // MQTT client instance
        private MqttClient client;

        #region ICommLayer ...

        public void Start(string deviceId, string deviceKey, string hostUri, string egressUri, string ingressUri, string cmdIngressUri)
        {
            this.client = new MqttClient(hostUri);
            this.client.MqttMsgPublishReceived += client_MqttMsgPublishReceived;
            this.client.Connect(deviceId, deviceId, deviceKey);
            this.client.Subscribe(new string[] { "out" }, new byte[] { MqttMsgBase.QOS_LEVEL_AT_LEAST_ONCE });
        }

        public void Stop()
        {
            this.client.Disconnect();
        }

        public void SendTelemetry(string deviceId, string subject, uint id, object value)
        {
            Message message = new Message()
            {
                Label = subject,
                Properties = 
                {
                    { "From", deviceId },
                    { "Time", DateTime.UtcNow },
                    { "dev", deviceId },
                    { "id", id },
                    { "val", value }
                }
            };

            this.client.Publish("in", message);
        }

        public void SendCommandReply(string deviceId, string gesture, string relatesTo, object result)
        {
            Message message = new Message()
            {
                Label = gesture,
                Properties = 
                {
                    { "From", deviceId },
                    { "Time", DateTime.UtcNow },
                    { "RelatesTo", relatesTo}, // correlation with command
                    { "res", result }
                }
            };

            this.client.Publish("in", message);
        }

        public event CommandReceivedCallback OnCommandReceived;

        #endregion

        void client_MqttMsgPublishReceived(object sender, MqttMsgPublishEventArgs e)
        {
#if (MF_FRAMEWORK_VERSION_V4_3)
            string jmessage = new string(Encoding.UTF8.GetChars(e.Message));
#else
            string jmessage = Encoding.UTF8.GetString(e.Message);
#endif

            try
            {
                Message message = Message.Parse(jmessage);

                if (message != null)
                {
                    IDictionary properties = new Hashtable();

                    // add gesture
                    properties.Add("gst", message.Label);
                    // add messageId
                    properties.Add("messageId", message.MessageId);

                    // add all other properties
                    foreach (DictionaryEntry entry in message.Properties)
                    {
                        properties.Add(entry.Key, entry.Value);
                    }

                    // raise command received event
                    CommandReceivedCallback callback = this.OnCommandReceived;

                    if (callback != null)
                        callback(this, properties);
                }
            }
            catch
            { }
        }

        class Message
        {
            public Message()
            {
                Properties = new Hashtable();
            }

            public static Message Parse(string jmessage)
            {
                Message m = null;

                try
                {
                    JObject jdom = (JObject)JsonHelpers.Parse(jmessage);

                    Hashtable properties = new Hashtable();

                    foreach (JProperty jprop in (JObject)jdom["p"])
                    {
                        properties.Add(jprop.Name, ((JValue)jprop.Value).RawValue);
                    }

                    m = new Message();
                    m.Properties = properties;
                }
                catch { }

                return m;
            }

            public string Label
            {
                get { return (string)Properties["subject"]; }
                set { Properties["subject"] = value; }
            }

            public string MessageId
            {
                get { return (string)Properties["message-id"]; }
                set { Properties["message-id"] = value; }
            }

            public IDictionary Properties { get; set; }


            public static implicit operator Byte[](Message m)
            {
                var jdom = new JObject();

                var jproperties = new JObject();

                foreach (DictionaryEntry entry in m.Properties)
                {
                    if (entry.Value.GetType() == typeof(double))
                        jproperties[(string)entry.Key] = (double)entry.Value;
                    else if (entry.Value.GetType() == typeof(int))
                        jproperties[(string)entry.Key] = (int)entry.Value;
                    else if (entry.Value.GetType() == typeof(uint))
                        jproperties[(string)entry.Key] = (uint)entry.Value;
                    else
                        jproperties[(string)entry.Key] = entry.Value.ToString();
                }

                jdom["p"] = jproperties;

                return Encoding.UTF8.GetBytes(JsonHelpers.Serialize(jdom));
            }
        }
    }
}
