﻿using System;
using System.Collections;
using IoT.Device.CommLayer;

namespace IoT.Device
{
    /// <summary>
    /// IoT Service to access Cloud
    /// </summary>
    public class IoTService
    {
        /// <summary>
        /// Washing start command delegate
        /// </summary>
        /// <param name="commandId">Command ID</param>
        /// <param name="program">Program</param>
        /// <param name="speed">Speed</param>
        /// <param name="temperature">Temperature</param>
        public delegate void WashingStartCallback(string commandId, int program, int speed, int temperature);

        /// <summary>
        /// Washing stop command delegate
        /// </summary>
        /// <param name="commandId">Command ID</param>
        public delegate void WashingStopCallback(string commandId);

        // subjects for telemetry/notification
        private string[] subjects = { "tmp", "hum", "pow", "not" };

        // device condifiguration
        private string deviceId;
        private string deviceKey;
        private string hostUri;
        private string ingressUri;
        private string egressUri;
        private string cmdIngressUri;

        // communication layer
        private ICommLayer commLayer;

        // events for washing command (on/off)
        public event WashingStartCallback OnWashingStart;
        public event WashingStopCallback OnWashingStop;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="commLayer"></param>
        public IoTService(ICommLayer commLayer)
        {
            // save reference to communication layer
            this.commLayer = commLayer;
            this.commLayer.OnCommandReceived += commLayer_OnCommandReceived;
        }

        /// <summary>
        /// Initialize washing machine system
        /// </summary>
        /// <param name="deviceId">Device identification</param>
        /// <param name="deviceKey">Device key</param>
        /// <param name="hostUri">Base host URI</param>
        /// <param name="ingressUri">Ingress (to cloud) URI for telemtry, inquiry</param>
        /// <param name="egressUri">Egress (from cloud) URI for command, notification</param>
        /// <param name="cmdIngressUri">Command ingress (to cloud) URI for command replies</param>
        public void Init(string deviceId, string deviceKey,
            string hostUri, string ingressUri, string egressUri, string cmdIngressUri)
        {
            this.deviceId = deviceId;
            this.deviceKey = deviceKey;
            this.hostUri = hostUri;
            this.ingressUri = ingressUri;
            this.egressUri = egressUri;
            this.cmdIngressUri = cmdIngressUri;
        }

        /// <summary>
        /// Start service
        /// </summary>
        public void Start()
        {
            // start communication layer with device configuration settings
            this.commLayer.Start(this.deviceId, 
                this.deviceKey, 
                this.hostUri, 
                this.egressUri, 
                this.ingressUri, 
                this.cmdIngressUri);
        }

        /// <summary>
        /// Stop service
        /// </summary>
        public void Stop()
        {
            // stop communication layer
            this.commLayer.Stop();
        }

        /// <summary>
        /// Send telemetry data / notification from device
        /// </summary>
        /// <param name="type">Type</param>
        /// <param name="id">Id (sensor instance)</param>
        /// <param name="value">Value</param>
        public void SendTelemetry(SubjectType type, uint id, object value)
        {
            string subject = this.subjects[(int)type];

            this.commLayer.SendTelemetry(this.deviceId, subject, id, value);
        }

        public void SendCommandReply(string gesture, string relatesTo, object result)
        {
            this.commLayer.SendCommandReply(this.deviceId, gesture, relatesTo, result);
        }

        bool commLayer_OnCommandReceived(object sender, IDictionary properties)
        {
            // command gesture not defined
            if (!properties.Contains("gst"))
                return false;

            switch ((string)properties["gst"])
            {
                case "wsh":
                    return this.Washing(properties);

                default:
                    return false;
            }
        }

        /// <summary>
        /// Execute washing command gesture
        /// </summary>
        /// <param name="properties">Command properties</param>
        /// <returns>Command result</returns>
        private bool Washing(IDictionary properties)
        {
            // "cmd" (on/off) not defined
            if (!properties.Contains("cmd"))
                return false;

            // start command (on)
            if (properties["cmd"].Equals("on"))
            {
                // check mandatory parameters
                if (!properties.Contains("prg") ||
                    !properties.Contains("spe") ||
                    !properties.Contains("tmp"))
                {
                    // wrong parameters
                    return false;
                }
                else
                {
                    try
                    {
                        // get parameters
                        int program = Convert.ToInt32(properties["prg"].ToString());
                        int speed = Convert.ToInt32(properties["spe"].ToString());
                        int temperature = Convert.ToInt32(properties["tmp"].ToString());

                        // raise event of application layer
                        WashingStartCallback callback = this.OnWashingStart;

                        if (callback != null)
                            callback((string)properties["messageId"], program, speed, temperature);

                        return true;
                    }
                    catch
                    {
                        return false;
                    }
                }
            }
            // stop command (off)
            else if (properties["cmd"].Equals("off"))
            {
                // raise event of application layer
                WashingStopCallback callback = this.OnWashingStop;

                if (callback != null)
                    callback((string)properties["messageId"]);

                return true;
            }
            // wrong command
            else
                return false;
        }

        /// <summary>
        /// Subject for telemetry
        /// </summary>
        public enum SubjectType
        {
            Temperature = 0,
            Humidity = 1,
            Power = 2,
            Notification = 3
        }
    }
}
