//  SensorLib.cpp : Defines the entry point for the console application.
//

#include "windows.h"
#include "windev.h"
#include "intlib.h"
#include "GpioLib.h"
#include "adc_ioctl.h"

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    return TRUE;
}

typedef struct
{
	DWORD pin;
	HANDLE intevt;
	DWORD sysintr;
} DHT11_INSTANCE;

void* WINAPI DHT11_Init(ULONG pin)
{
	DHT11_INSTANCE* inst;
	DWORD			irq;
	PIN_INSTANCE	gpio;

	inst = (DHT11_INSTANCE*)malloc(sizeof(DHT11_INSTANCE));

	if (!inst)
		return NULL;

	memset(inst, 0, sizeof(DHT11_INSTANCE));

	inst->intevt = CreateEvent(NULL, FALSE, FALSE, NULL);

	if (inst->intevt == NULL)
		return NULL;

	if (!GetGPIOFromPin(pin, FALSE, &gpio))
		return NULL;

	if (!SetGPIOIrqEdge(gpio.inst1, GPIO_EDGE_FALLING))
		return NULL;

	irq = GetGPIOIrq(gpio.inst1);
	inst->sysintr = RequestSysInterrupt(irq);

	if (!InterruptInitializeCompat(inst->sysintr, inst->intevt, NULL, 0))
		return NULL;

	// configure pin as output (should keep it high)
	SetPinLevel(pin, 1);
	SetPinAltFn(pin, 0, TRUE);
	SetPinDir(pin, TRUE);

	inst->pin = pin;

	return inst;
}


BOOL WINAPI DHT11_Read(void* instance, double* humidity, double* temperature)
{
	DHT11_INSTANCE* inst;
	DWORD priority;
	LARGE_INTEGER	freq;
	BYTE			data[5];
	DWORD			index = 0;
	DWORD			start = 0;
	LARGE_INTEGER	prev;
	LARGE_INTEGER	curr;
	LARGE_INTEGER	threshold;

	OutputDebugString(TEXT("Data read.\r\n"));

	inst = (DHT11_INSTANCE*)instance;

	if ((inst == NULL) || (humidity == NULL) || (temperature == NULL))
	{
		OutputDebugString(TEXT("Invalid pointers.\r\n"));
		return FALSE;
	}

	priority = CeSetThreadPriority(GetCurrentThread(), 10);

	memset(data, 0, 5);

	QueryPerformanceFrequency(&freq);

	// pulse for 0 is 23-24uS, pulse for 1 is 71us positive pulse plus 54us down, so we use 105us as threshold value
	threshold.QuadPart = freq.QuadPart * 105 / 1000000;

	curr.QuadPart = 0;

	// generate a 1ms low pulse
	SetPinLevel(inst->pin, 1); 
	Sleep(1);
	SetPinLevel(inst->pin, 0);
	Sleep(1);
	SetPinLevel(inst->pin, 1);

	InterruptDoneCompat(inst->sysintr); // cleans pending interrupt
	SetPinDir(inst->pin, FALSE);

	// the sensors usually replies in 10ms... or doesn't reply at all
	while (WaitForSingleObject(inst->intevt, 50) == WAIT_OBJECT_0)
	{
		// falling edge
		prev.QuadPart = curr.QuadPart;
		QueryPerformanceCounter(&curr);

		// first we have to wait until start had been received
		if (start == 2)
		{
			if (curr.QuadPart - prev.QuadPart > threshold.QuadPart)
				data[index >> 3] |= 1 << (7 - (index & 0x07));

			index++;
			if (index == 40)	// we completed the read
			{
				InterruptDoneCompat(inst->sysintr);
				break;
			}
		}
		else
		{
			start++;
		}

		InterruptDoneCompat(inst->sysintr);
	}

	CeSetThreadPriority(GetCurrentThread(), priority);

	SetPinDir(inst->pin, TRUE);
	SetPinLevel(inst->pin, 1);

	// check that we received the full data packet
	if (index < 40)
	{
		TCHAR buffer[255];

		_stprintf(buffer, TEXT("No data - %d\r\n"), index);
		OutputDebugString(buffer);
		return FALSE;
	}

	// validate checksum
	if (data[4] != (BYTE)(data[0] + data[1] + data[2] + data[3]))
	{
		TCHAR buffer[255];

		_stprintf(buffer, TEXT("Invalid data - %d %d %d %d %d\r\n"), (int)data[0],(int)data[1],(int)data[2],(int)data[3],(int)data[4]);
		OutputDebugString(buffer);
		return FALSE;
	}

	*humidity = data[0];
	*temperature = data[2];

	OutputDebugString(TEXT("OK\r\n"));
	return TRUE;
}

void WINAPI DHT11_Release(void* instance)
{
	DHT11_INSTANCE* inst;

	inst = (DHT11_INSTANCE*)instance;

	if (inst->sysintr != 0xFFFFFFFF)
	{
		InterruptDisableCompat(inst->sysintr);
		ReleaseSysIntr(inst->sysintr);
	}

	free(inst);
}

typedef struct
{
	DWORD	pins[4];
	HANDLE	stepevents[2];
	HANDLE	stepthread;
} Stepper_INSTANCE;

DWORD sequence[4][4] = { { 1, 1, 0, 0 }, { 0, 1, 1, 0 }, { 0, 0, 1, 1 }, { 1, 0, 0, 1 } };

DWORD WINAPI Stepper_Thread(void* instptr)
{
	Stepper_INSTANCE*	instance;
	DWORD				waitres;
	int					i,j;

	instance = (Stepper_INSTANCE*)instptr;

	CeSetThreadPriority(GetCurrentThread(), 20);

	while ((waitres = WaitForMultipleObjects(3, instance->stepevents, FALSE, INFINITE)) == WAIT_OBJECT_0+1)
	{
		SleepTillTick();

		for (i = 0 ; i < 4 ; i ++)
		{
			// we need to generate a pulse on each pin in sequence
			for (j = 0; j < 4; j++)
				SetPinLevel(instance->pins[j], sequence[i][j]);
			Sleep(1);
		}

		Sleep(1);
	}
	return 0;
}

void* WINAPI Stepper_Init(DWORD pin0, DWORD pin1, DWORD pin2, DWORD pin3)
{
	Stepper_INSTANCE* instance;
	int i;

	instance = (Stepper_INSTANCE*)malloc(sizeof(Stepper_INSTANCE));

	if (!instance)
		return NULL;

	memset(instance, 0, sizeof(Stepper_INSTANCE));

	instance->pins[0] = pin0;
	instance->pins[1] = pin1;
	instance->pins[2] = pin2;
	instance->pins[3] = pin3;

	for (i = 0; i < 4; i++)
	{
		SetPinLevel(instance->pins[i], 0);
		SetPinAltFn(instance->pins[i], 0, TRUE);
		SetPinDir(instance->pins[i], TRUE);
	}

	// 3 events: quit, rotate clockwise, rotate counterclockwise
	instance->stepevents[0] = CreateEvent(NULL, TRUE, FALSE, NULL);
	instance->stepevents[1] = CreateEvent(NULL, TRUE, FALSE, NULL);

	if (!instance->stepevents[0] || !instance->stepevents[1] )
		goto failed;

	instance->stepthread = CreateThread(NULL, 0, Stepper_Thread, instance, 0, NULL);

	if (!instance->stepthread)
		goto failed;

	return instance;

failed:

	CloseHandle(instance->stepevents[0]);
	CloseHandle(instance->stepevents[1]);

	free(instance);
	return NULL;
}

void WINAPI Stepper_Start(void* inst)
{
	Stepper_INSTANCE* instance = (Stepper_INSTANCE*)inst;

	SetEvent(instance->stepevents[1]);
}

void WINAPI Stepper_Stop(void* inst)
{
	int i;
	Stepper_INSTANCE* instance = (Stepper_INSTANCE*)inst;

	ResetEvent(instance->stepevents[1]);

	// wait until current step has been completed
	Sleep(100);

	for (i = 0; i < 4;i++)
		SetPinLevel(instance->pins[i], 0);
}

void WINAPI Stepper_Release(void* inst)
{
	int i;
	Stepper_INSTANCE* instance = (Stepper_INSTANCE*)inst;

	SetEvent(instance->stepevents[0]);
	ResetEvent(instance->stepevents[1]);
	WaitForSingleObject(instance->stepthread, INFINITE);

	for (i = 0; i < 4; i++)
		SetPinLevel(instance->pins[i], 0);

	CloseHandle(instance->stepthread);
	CloseHandle(instance->stepevents[0]);
	CloseHandle(instance->stepevents[1]);

	free(instance);
}

#define ADC_PINS_COUNT 4

typedef struct 
{
	DWORD pin;
	DWORD af;
	DWORD adc;
	DWORD channel;
} ADC_PIN;

HANDLE adchandle[2] = { INVALID_HANDLE_VALUE, INVALID_HANDLE_VALUE };
ADC_PIN adcpins[ADC_PINS_COUNT] = { { 8, 1, 0, 8 }, { 6, 1, 1, 8 }, { 4, 1, 0, 9 }, { 2, 1, 1, 9 } };

BOOL WINAPI ADC_Init()
{
	DWORD i;

	adchandle[0] = CreateFile(TEXT("ADC0:"), GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);

	if (adchandle[0] == INVALID_HANDLE_VALUE)
		return FALSE;

	adchandle[1] = CreateFile(TEXT("ADC1:"), GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);

	if (adchandle[1] == INVALID_HANDLE_VALUE)
		return FALSE;

	for (i = 0; i < ADC_PINS_COUNT; i++)
	{
		SetPinAltFn(adcpins[i].pin, adcpins[i].af,FALSE);
	}
	return TRUE;
}

void WINAPI ADC_Release()
{
	if (adchandle[0] != INVALID_HANDLE_VALUE)
	{
		CloseHandle(adchandle[0]);
		adchandle[0] = INVALID_HANDLE_VALUE;
	}

	if (adchandle[1] != INVALID_HANDLE_VALUE)
	{
		CloseHandle(adchandle[1]);
		adchandle[1] = INVALID_HANDLE_VALUE;
	}
}

DWORD WINAPI ADC_Read(DWORD channel)
{
	DWORD value;
	DWORD returned;
	ADC_READ adcread;

	if (channel >= ADC_PINS_COUNT)
		return (DWORD)-1;

	adcread.channel = adcpins[channel].channel;
	adcread.mode = ADC_MODE_12BIT;
	adcread.useaverage = TRUE;
	adcread.avgsamples = ADC_AVG_4;

	if (!DeviceIoControl(adchandle[adcpins[channel].adc], IOCTL_ADC_READ, &adcread, sizeof(ADC_READ), &value, sizeof(DWORD), &returned, NULL))
	{
		return (DWORD)-1;
	}

	return value;
}