﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Threading;
using SensorLib.NetCf39;
using IoT.Device.CommLayer;
using IoT.Device;

namespace CDays2014App
{
    public partial class WashingMachineForm : Form
    {
        // device identification (ID and Key)
        private string deviceId = "<deviceId>";
        private string deviceKey = "<deviceKey>";
        private string hostUri = "<hostUri>";
        // ingress and egress URI for messages exchanges
        private string egressUri = "/out/t0002/Subscriptions/s0000";
        private string ingressUri = "/in/t000A";
        private string cmdIngressUri = "/in/t00A";

        // IoT Service for sending data and receiving commands
        private IoTService iotService;

        // Pin used for the DHT temperature sensor
        private const uint DHT11_PIN_NUM = 45;

        // Pins used to control the stepper motor 
        private readonly uint[] STEPPER_PINS = new uint[] { 98, 103, 97, 79 };

        // classes used for sensors and stepper motor control
        private DHT11 dht11=null;
        private Stepper stepper = null;
        private ADC adc = null;

        // Array used to store the values read from multiple ADC inputs
        private Label[] powerLabels = null;

        public WashingMachineForm()
        {
            InitializeComponent();

            powerLabels = new Label[3];
            powerLabels[0] = adc0value;
            powerLabels[1] = adc1value;
            powerLabels[2] = adc2value;
        }

        private void WashingMachineForm_Load(object sender, EventArgs e)
        {
            dht11 = new DHT11(DHT11_PIN_NUM);
            stepper = new Stepper(STEPPER_PINS);
            adc = new ADC();

            DHT11Timer.Enabled = true;
            ADCTimer.Enabled = true;

            // create IoTService instance with AMQP communication layer
            this.iotService = new IoTService(new AmqpCommLayer());

            this.iotService.Init(deviceId,
                deviceKey,
                hostUri,
                ingressUri,
                egressUri,
                cmdIngressUri);

            // register to washing commands
            this.iotService.OnWashingStart += iotService_OnWashingStart;
            this.iotService.OnWashingStop += iotService_OnWashingStop;

            // start service
            this.iotService.Start();
        }

        private void WashingMachineForm_Closed(object sender, EventArgs e)
        {
            dht11.Release();
            stepper.Release();
            adc.Release();
            iotService.Stop();
        }

        
        // variables used to store current values 
        private double currentHumidity = double.MinValue;
        private double currentTemperature = double.MinValue;

        private void DHT11Timer_Tick(object sender, EventArgs e)
        {
            double humidity = 0.0, temperature = 0.0;
    
            try
            {
                if (!dht11.Read(out humidity, out temperature))
                    return; // sometimes the DHT sensor does not reply

                if (currentTemperature!=temperature)
                {
                    if (sendData.Checked)
                    {
                        iotService.SendTelemetry(IoTService.SubjectType.Temperature, 0, temperature);
                        currentTemperature = temperature;
                    }
                    tempLabel.Text = temperature.ToString()+"C°";
                }
                if (currentHumidity != humidity)
                {
                    if (sendData.Checked)
                    {
                        iotService.SendTelemetry(IoTService.SubjectType.Humidity, 0, humidity);
                        currentHumidity = humidity;
                    }    
                    humidityLabel.Text = humidity.ToString()+"%";
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.ToString());
            }
        }

        private void spinClockwise_Click(object sender, EventArgs e)
        {
            stepper.Start();
        }


        private void stopSpinning_Click(object sender, EventArgs e)
        {
            stepper.Stop();
        }

        private const int POWER_SENSORS_COUNT = 3;

        private double[] powerConsumption = new double[] { double.MinValue, double.MinValue, double.MinValue };
        
        // we need different values because the circuit connected to ADC inputs use different diodes
        private double[] factors = new double[] { 17.32, 4.62, 17.32 };

        private void ADCTimer_Tick(object sender, EventArgs e)
        {
            try
            {
                for (uint i=0;i<POWER_SENSORS_COUNT;i++)
                {
                    double power;

                    // we need to convert the value from the ADC to a value in W
                    power = Math.Round(((double)adc.Read(i))/factors[i]);
                    
                    if (power!=powerConsumption[i])
                    {
                        if (sendData.Checked)
                        {
                            iotService.SendTelemetry(IoTService.SubjectType.Power, i, power);
                            powerConsumption[i] = power;
                        }
                        powerLabels[i].Text = power.ToString()+"W";
                    }
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.ToString());
            }
        }

        void iotService_OnWashingStop(string commandId)
        {
            try
            {
                stepper.Stop();
                this.iotService.SendCommandReply("wsh", commandId, "ok");
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.ToString());
            }
        }

        System.Threading.Timer timer = null;

        void iotService_OnWashingStart(string commandId, int program, int speed, int temperature)
        {
            if (timer!=null)
                return;

            try
            {
                stepper.Start();
               timer = new System.Threading.Timer(WashingTimer_Tick,null,10000,Timeout.Infinite);                

                this.iotService.SendCommandReply("wsh", commandId, "ok");
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.ToString());
            }
        }

        private void WashingTimer_Tick(object sender)
        {
            try
            {
                timer.Change(Timeout.Infinite, Timeout.Infinite);
                timer.Dispose();
                timer = null;

                stepper.Stop();
                this.iotService.SendTelemetry(IoTService.SubjectType.Notification, 0, "ok");
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.ToString());
            }

        }

        private void sendData_CheckStateChanged(object sender, EventArgs e)
        {
            // we reset previous values to force a re-send at the next update
            if (sendData.Checked)
            {
                currentHumidity = double.MinValue;
                currentTemperature = double.MinValue;
                for (int i = 0; i < POWER_SENSORS_COUNT; i++)
                    powerConsumption[i] = double.MinValue;
            }
        }
    }
}
